"""
Simplemonitor logger for MQTT

It is intended to be used with Home Assistant and its MQTT Discovery feature
(this the default topic)
but can be used in any other context
contact: dev@swtk.info
"""

import json
from typing import List, Optional, cast

import paho.mqtt.publish

from ..Monitors.monitor import Monitor
from .logger import Logger, register


@register
class MQTTLogger(Logger):
    """Log to MQTT endpoints"""

    logger_type = "mqtt"
    only_failures = False
    buffered = False
    dateformat = None

    def __init__(self, config_options: Optional[dict] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)

        self.host = cast(
            str, self.get_config_option("host", required=True, allow_empty=False)
        )
        self.port = cast(
            int,
            self.get_config_option(
                "port",
                required=False,
                allow_empty=False,
                required_type="int",
                default=1883,
            ),
        )
        # specif configuration for Home Assistant MQTT discovery
        # https://www.home-assistant.io/docs/mqtt/discovery/
        self.hass = cast(
            bool,
            self.get_config_option(
                "hass",
                required=False,
                allow_empty=False,
                required_type="bool",
                default=False,
            ),
        )
        # topic to send information to
        self.topic = cast(
            str,
            self.get_config_option(
                "topic",
                required=False,
                allow_empty=False,
                default=(
                    "simplemonitor" if not self.hass else "homeassistant/binary_sensor"
                ),
            ),
        )
        # username for authentication
        self.username = cast(
            str,
            self.get_config_option(
                "username",
                required=False,
                allow_empty=True,
            ),
        )
        # password for authentication
        self.password = cast(
            str,
            self.get_config_option(
                "password",
                required=False,
                allow_empty=True,
            ),
        )
        self.device_class = cast(
            str,
            self.get_config_option("device_class", default=""),
        )

        if self.username and self.password:
            self.auth = {"username": self.username, "password": self.password}
        else:
            self.auth = {}

        # registry of monitors which registered with HA
        # not used if not Home Assistant context
        # also see
        # https://github.com/jamesoff/simplemonitor/issues/236#issuecomment-462481900
        # for rationale
        self.registered = []  # type: List[str]

    def save_result2(self, name: str, monitor: Monitor) -> None:
        safe_name = monitor.name
        if self.hass:
            if " " in safe_name:
                self.logger_logger.warning(
                    (
                        "replacing spaces with underscores for monitor %s as spaces are "
                        "not supported for MQTT/HASS names"
                    ),
                    monitor.name,
                )
                safe_name = monitor.name.replace(" ", "_")
            if monitor.name not in self.registered:
                self.logger_logger.info(
                    "attempting to register MQTT config topic for monitor %s", name
                )
                config_payload = {
                    "name": monitor.name,
                    "state_topic": "{root}/simplemonitor_{monitor}/state".format(
                        root=self.topic, monitor=safe_name
                    ),
                }
                if self.device_class:
                    config_payload["device_class"] = self.device_class
                try:
                    paho.mqtt.publish.single(
                        "{root}/simplemonitor_{monitor}/config".format(
                            root=self.topic,
                            monitor=safe_name,
                        ),
                        payload=json.dumps(config_payload),
                        retain=True,
                        hostname=self.host,
                        port=self.port,
                        auth=self.auth,
                        client_id="simplemonitor_{monitor}".format(
                            monitor=safe_name,
                        ),
                    )
                except Exception:
                    self.logger_logger.exception("cannot send %s to MQTT", monitor.name)
                else:
                    self.registered.append(monitor.name)
                    self.logger_logger.debug("registered %s in MQTT", safe_name)

        if self.only_failures and monitor.virtual_fail_count() == 0:
            return

        if self.hass:
            topic = "{root}/simplemonitor_{monitor}/state".format(
                root=self.topic, monitor=safe_name
            )
        else:
            topic = "{root}/{monitor}".format(root=self.topic, monitor=safe_name)
        self.logger_logger.debug(
            "%s failed %d times", monitor.name, monitor.virtual_fail_count()
        )
        payload = (
            "ON"
            if monitor.virtual_fail_count() == 0 and not monitor.was_skipped
            else "OFF"
        )
        try:
            paho.mqtt.publish.single(
                topic,
                payload=payload,
                retain=True,
                hostname=self.host,
                port=self.port,
                auth=self.auth,
                client_id="simplemonitor_{monitor}".format(monitor=safe_name),
            )
        except Exception:
            self.logger_logger.exception("cannot send state %s to %s", payload, topic)
        else:
            self.logger_logger.debug("state %s sent to %s", payload, topic)

    def describe(self) -> str:
        return "Sends monitoring status to a MQTT broker"



"""
SimpleMonitor alerting via BulkSMS
"""

from typing import cast

import requests

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class BulkSMSAlerter(Alerter):
    """
    Send SMS alerts using the BulkSMS service

    Subscription required, see http://www.bulksms.co.uk"""

    alerter_type = "bulksms"
    urgent = True

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.username = cast(
            str, self.get_config_option("username", required=True, allow_empty=False)
        )
        self.password = cast(
            str, self.get_config_option("password", required=True, allow_empty=False)
        )
        self.target = cast(
            str, self.get_config_option("target", required=True, allow_empty=False)
        )

        self.sender = cast(str, self.get_config_option("sender", default="SmplMntr"))
        if len(self.sender) > 11:
            self.alerter_logger.warning("truncating SMS sender name to 11 chars")
            self.sender = self.sender[:11]

        self.api_host = self.get_config_option("api_host", default="www.bulksms.co.uk")
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

        self.support_catchup = True

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Send an SMS alert."""

        alert_type = self.should_alert(monitor)
        if alert_type not in [AlertType.FAILURE, AlertType.SUCCESS]:
            return

        message = self.build_message(AlertLength.SMS, alert_type, monitor)

        url = f"https://{self.api_host}/eapi/submission/send_sms/2/2.0"
        params = {
            "username": self.username,
            "password": self.password,
            "message": message,
            "msisdn": self.target,
            "sender": self.sender,
            "repliable": "0",
        }

        if not self._dry_run:
            try:
                response = requests.get(url, params=params, timeout=self.timeout)
                status = response.text
                if not status.startswith("0"):
                    self.alerter_logger.error(
                        "Unable to send SMS: %s (%s)",
                        status.split("|")[0],
                        status.split("|")[1],
                    )
            except requests.exceptions.RequestException:
                self.alerter_logger.exception("SMS sending failed")
        else:
            self.alerter_logger.info(
                "dry_run: would send SMS: %s with message %s", url, message
            )

    def _describe_action(self) -> str:
        return "SMSing {target} via BulkSMS".format(target=self.target)



"""Utilities for SimpleMonitor."""

import datetime
import os
import shutil
import socket
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import arrow

from .envconfig import EnvironmentAwareConfigParser


class MonitorConfigurationError(ValueError):
    """A config error for a Monitor"""


class AlerterConfigurationError(ValueError):
    """A config error for an Alerter"""


class LoggerConfigurationError(ValueError):
    """A config error for a Logger"""


class SimpleMonitorConfigurationError(ValueError):
    """A general config error"""


class MonitorState(Enum):
    """Represent the state of a Monitor."""

    UNKNOWN = 0  # state not known yet
    SKIPPED = 1  # monitor was skipped
    OK = 2  # monitor is ok
    FAILED = 3  # monitor has failed


class UpDownTime:
    """Represent an up- or downtime"""

    days = 0
    hours = 0
    minutes = 0
    seconds = 0

    def __init__(
        self, days: int = 0, hours: int = 0, minutes: int = 0, seconds: int = 0
    ) -> None:
        if not isinstance(days, int):
            raise TypeError("days must be an int")
        if not isinstance(hours, int):
            raise TypeError("days must be an int")
        if not isinstance(minutes, int):
            raise TypeError("days must be an int")
        if not isinstance(seconds, int):
            raise TypeError("days must be an int")
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds
        if self.seconds >= 60:
            temp_min, self.seconds = divmod(self.seconds, 60)
            self.minutes += temp_min
        if self.minutes >= 60:
            temp_hour, self.minutes = divmod(self.minutes, 60)
            self.hours += temp_hour
        if self.hours >= 24:
            temp_day, self.hours = divmod(self.hours, 24)
            self.days += temp_day

    def __str__(self) -> str:
        """Format as d+h:m:s"""
        return "{}+{:02}:{:02}:{:02}".format(
            self.days, self.hours, self.minutes, int(self.seconds)
        )

    def __repr__(self) -> str:
        return "<{}: {}>".format(self.__class__, self.__str__())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UpDownTime):
            return NotImplemented
        if (
            self.days == other.days
            and self.hours == other.hours
            and self.minutes == other.minutes
            and self.seconds == other.seconds
        ):
            return True
        return False

    @staticmethod
    def from_timedelta(td: datetime.timedelta) -> "UpDownTime":
        """Generate an UpDownTime from a timedelta object"""
        if td is None:
            return UpDownTime()
        downtime_seconds = td.seconds
        (hours, minutes) = (0, 0)
        if downtime_seconds > 3600:
            (hours, downtime_seconds) = divmod(downtime_seconds, 3600)
        if downtime_seconds > 60:
            (minutes, downtime_seconds) = divmod(downtime_seconds, 60)
        return UpDownTime(td.days, hours, minutes, downtime_seconds)


def get_config_option(
    config_options: Dict[str, Any],
    key: str,
    *,
    default: Any = None,
    required: bool = False,
    required_type: str = "str",
    allowed_values: Any = None,
    allow_empty: bool = True,
    minimum: Optional[Union[int, float]] = None,
    maximum: Optional[Union[int, float]] = None,
) -> Union[None, str, int, float, bool, List[str], List[int]]:
    """Get a value out of a dict, with possible default, required type and requiredness."""

    if not isinstance(config_options, dict):
        raise TypeError("config_options should be a dict")

    value = config_options.get(key, default)
    if required and value is None:
        raise ValueError(f"config option {key} is missing and is required")
    if isinstance(value, str) and required_type:
        if required_type == "str" and value == "" and not allow_empty:
            raise ValueError(f"config option {key} cannot be empty")
        if required_type in ["int", "float"]:
            try:
                if required_type == "int":
                    value = int(value)
                else:
                    value = float(value)
            except TypeError as error:
                raise TypeError(
                    f"config option {key} needs to be an {required_type}"
                ) from error
            if minimum is not None and value < minimum:
                raise ValueError(f"config option {key} needs to be >= {minimum}")
            if maximum is not None and value > maximum:
                raise ValueError(f"config option {key} needs to be <= {minimum}")
        if required_type == "[int]":
            if not isinstance(value, str):
                raise ValueError(
                    f"config option {key} needs to be a list of int,int,..."
                )
            try:
                value = [int(x) for x in value.split(",")]
            except ValueError as error:
                raise ValueError(
                    f"config option {key} needs to be a list of int[int,...]"
                ) from error
        if required_type == "bool":
            value = bool(str(value).lower() in ["1", "true", "yes"])
        if required_type == "[str]":
            if not isinstance(value, str):
                raise ValueError(
                    f"config option {key} needs to be a list of int,int,..."
                )
            value = [x.strip() for x in value.split(",")]
    if isinstance(value, list) and allowed_values:
        if not all(x in allowed_values for x in value):
            raise ValueError(f"config option {key} needs to be one of {allowed_values}")
    else:
        if allowed_values is not None and value not in allowed_values:
            raise ValueError(f"config option {key} needs to be one of {allowed_values}")
    return value


def format_datetime(
    the_datetime: Optional[Union[arrow.Arrow, datetime.datetime]],
    tz: Optional[str] = None,
) -> str:
    """Return an isoformat()-like datetime without the microseconds."""
    if the_datetime is None:
        retval = ""
    elif isinstance(the_datetime, (arrow.Arrow, datetime.datetime)):
        the_datetime = the_datetime.replace(microsecond=0)
        retval = the_datetime.isoformat(" ")
        if isinstance(the_datetime, arrow.Arrow):
            if tz is not None:
                retval = the_datetime.to(tz).isoformat(" ")
    else:
        retval = str(the_datetime)
    return retval


def short_hostname() -> str:
    """Get just our machine name.

    TODO: This might actually be redundant. Python probably provides it's own version of this.
    """

    return (socket.gethostname() + ".").split(".")[0]


def get_config_dict(
    config: EnvironmentAwareConfigParser, monitor: str
) -> Dict[str, str]:
    options = config.items(monitor)
    ret = {}
    for key, value in options:
        ret[key] = value
    return ret


def subclass_dict_handler(
    mod: str, base_cls: type, type_attr: str
) -> Tuple[Callable, Callable, Callable]:
    def _check_is_subclass(cls: Any) -> None:
        if not issubclass(cls, base_cls):
            raise TypeError(
                ("%s.register may only be used on subclasses of %s.%s")
                % (mod, mod, base_cls.__name__)
            )

    _subclasses = {}

    def register(cls: Any) -> Any:
        """Decorator for monitor classes."""
        _check_is_subclass(cls)
        if cls is None or getattr(cls, type_attr, "unknown") == "unknown":
            raise ValueError("Cannot register this class")
        _subclasses[getattr(cls, type_attr)] = cls
        return cls

    def get_class(type_: Any) -> Any:
        return _subclasses[type_]

    def all_types() -> list:
        return list(_subclasses)

    return (register, get_class, all_types)


def check_group_match(group: Union[str, List[str]], group_list: List[str]) -> bool:
    """
    Check if a group is contained in the group list.

    If the group list is a single element, "_all", then it matches.
    """
    if group_list[0] == "_all":
        return True

    if isinstance(group, str):
        group = group.strip().split(",")

    if any(element in group_list for element in group):
        return True
    return False


def size_string_to_bytes(s: str) -> Optional[int]:
    if s is None:
        return None
    if s.endswith("G"):
        gigs = int(s[:-1])
        _bytes = gigs * (1024**3)
    elif s.endswith("M"):
        megs = int(s[:-1])
        _bytes = megs * (1024**2)
    elif s.endswith("K"):
        kilos = int(s[:-1])
        _bytes = kilos * 1024
    else:
        return int(s)
    return _bytes


def bytes_to_size_string(b: int) -> str:
    """Convert a number in bytes to a sensible unit."""

    kb = 1024
    mb = kb * 1024
    gb = mb * 1024
    tb = gb * 1024

    if b > tb:
        return "%0.2fTiB" % (b / float(tb))
    if b > gb:
        return "%0.2fGiB" % (b / float(gb))
    if b > mb:
        return "%0.2fMiB" % (b / float(mb))
    if b > kb:
        return "%0.2fKiB" % (b / float(kb))
    return str(b)


def copy_if_different(source: str, dest: str) -> bool:
    """Copy a file from src to dest, if newer or a different size"""
    do_copy = False
    if not os.path.exists(source):
        return False
    if os.path.isdir(dest):
        dest = os.path.join(dest, os.path.basename(source))
    if not os.path.exists(dest):
        do_copy = True
    else:
        source_fileinfo = os.stat(source)
        dest_fileinfo = os.stat(dest)
        if source_fileinfo.st_size != dest_fileinfo.st_size:
            do_copy = True
        elif source_fileinfo.st_mtime > dest_fileinfo.st_mtime:
            do_copy = True
    if not do_copy:
        return False
    try:
        shutil.copy(source, dest)
    except IOError:
        return False
    return True



"""
Service monitoring for SimpleMonitor
"""

import fnmatch
import os
import platform
import re
import subprocess
import time
from typing import Any, List, Optional, Tuple, Union, cast

import psutil

from .monitor import Monitor, register

try:
    import pydbus
except ImportError:
    pydbus = None


@register
class MonitorSvc(Monitor):
    """Monitor a service handled by daemontools."""

    monitor_type = "svc"
    path = ""
    regexp = re.compile(r": up \(pid \d+\) (?P<uptime>\d+) seconds")

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.path = cast(str, self.get_config_option("path", required=True))
        self.params = ["svstat", self.path]
        self.minimum_uptime = int(
            self.get_config_option(
                "minimum_uptime", required_type="int", minimum=0, default=0
            )
        )

    def run_test(self) -> bool:
        if self.path == "":
            return self.record_fail("Path is not configured")
        try:
            result = subprocess.check_output(self.params).decode()
            matches = self.regexp.search(result)
            if not matches:
                return self.record_fail(result)

        except Exception as error:
            return self.record_fail(f"Exception while executing svstat: {error!r}")
        try:
            uptime = int(matches.group("uptime"))
            if uptime < self.minimum_uptime:
                return self.record_fail(
                    f"{result} (wanted {self.minimum_uptime} seconds)"
                )
            return self.record_success(result)
        except ValueError:
            return self.record_fail("Could not parse output from svstat")

    def describe(self) -> str:
        return f"Checking that the supervise-managed service in {self.path} is up for at least {self.minimum_uptime} seconds."

    def get_params(self) -> Tuple:
        return (self.path, self.uptime)


@register
class MonitorService(Monitor):
    """Monitor a Windows service"""

    service_name = ""
    want_state = "RUNNING"
    host = "."
    monitor_type = "service"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if psutil is None:
            self.monitor_logger.critical("psutil is not installed.")
            self.monitor_logger.critical("Try: pip install -r requirements.txt")
        self.service_name = cast(str, self.get_config_option("service", required=True))
        self.want_state = cast(
            str,
            self.get_config_option(
                "state",
                default="RUNNING",
                allowed_values=[
                    "RUNNING",
                    "STOPPED",
                    "PAUSED",
                    "START_PENDING",
                    "PAUSE_PENDING",
                    "CONTINUE_PENDING",
                    "STOP_PENDING",
                ],
            ),
        )
        self.host = cast(str, self.get_config_option("host", default="."))

    def run_test(self) -> bool:
        """Check the service is in the desired state"""
        if psutil is None:
            return self.record_fail("psutil is not installed")
        try:
            service = psutil.win_service_get(self.service_name)
        except psutil.NoSuchProcess:
            return self.record_fail(
                "service {} does not exist".format(self.service_name)
            )
        except AttributeError:
            return self.record_fail("not supported on this platform")
        except Exception:
            self.monitor_logger.exception("Failed to get service")
            return self.record_fail("Unable to get service")

        _state = service.status()
        if _state:
            state = _state.upper()
        else:
            state = "NONE"
        if state != self.want_state:
            return self.record_fail(
                "Service state is {} (wanted {})".format(state, self.want_state)
            )
        return self.record_success()

    def describe(self) -> str:
        """Explains what this instance is checking"""
        return "checking for service called %s in state %s" % (
            self.service_name,
            self.want_state,
        )

    def get_params(self) -> Tuple:
        return (self.host, self.service_name, self.want_state)


@register
class MonitorRC(Monitor):
    """Monitor a service handled by an rc.d script.

    This monitor checks the return code of /usr/local/etc/rc.d/<name>
    and reports failure if it's non-zero by default.
    """

    monitor_type = "rc"

    def __init__(self, name: str, config_options: dict) -> None:
        """Initialise the class.
        Change script path to /etc/rc.d/ to monitor base system services. If the
        script path ends with /, the service name is appended."""
        super().__init__(name, config_options)
        self.service_name = cast(str, self.get_config_option("service", required=True))
        self.script_path = cast(
            str, self.get_config_option("path", default="/usr/local/etc/rc.d/")
        )
        self.want_return_code = cast(
            str, self.get_config_option("return_code", required_type="int", default=0)
        )
        if self.script_path.endswith("/"):
            self.script_path = self.script_path + self.service_name
        if not os.path.isfile(self.script_path):
            if os.path.isfile(self.script_path + ".sh"):
                self.script_path = self.script_path + ".sh"
            else:
                raise RuntimeError("Script %s(.sh) does not exist" % self.script_path)

    def run_test(self) -> bool:
        """Check the service is in the desired state."""
        if platform.system() in ["Microsoft", "CYGWIN_NT-6.0"]:
            return self.record_fail("Cannot run this monitor on a non-UNIX host.")
        try:
            returncode = subprocess.check_call([self.script_path, "status"])
            if returncode == self.want_return_code:
                return self.record_success()
        except subprocess.CalledProcessError as error:
            if error.returncode == self.want_return_code:
                return self.record_success()
            returncode = -1
        except Exception as error:
            return self.record_fail("Exception while executing script: %s" % error)
        return self.record_fail(
            "Return code: %d (wanted %d)" % (returncode, int(self.want_return_code))
        )

    def get_params(self) -> Tuple:
        return (self.service_name, self.want_return_code)

    def describe(self) -> str:
        """Explains what this instance is checking."""
        return "Checks service %s is running" % self.script_path


@register
class MonitorUnixService(Monitor):
    """Monitor a service handled by a generic "service" command.

    If "service X status" exits 0 for the service being up, and non-zero
    otherwise, this is for you.
    """

    monitor_type = "unix_service"

    def __init__(
        self, name: str = "unnamed", config_options: Optional[dict] = None
    ) -> None:
        super().__init__(name=name, config_options=config_options)
        self.service_name = cast(str, self.get_config_option("service", required=True))
        self.want_state = cast(
            str,
            self.get_config_option(
                "state", allowed_values=["running", "stopped"], default="running"
            ),
        )
        if self.want_state == "running":
            self._want_return_code = 0
        else:
            self._want_return_code = 1
        self.jail = cast(Union[str, None], self.get_config_option("jail"))

    def run_test(self) -> bool:
        try:
            result = subprocess.run(
                [
                    "service",
                    *(["-j", self.jail] if self.jail else []),
                    self.service_name,
                    "status",
                ],
                check=False,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )  # nosec
            returncode = result.returncode
        except subprocess.SubprocessError as exception:
            return self.record_fail(
                "Failed to run 'service {} status: {}".format(
                    self.service_name, exception
                )
            )
        if returncode == self._want_return_code:
            return self.record_success()
        return self.record_fail(
            "Got exit code {}, wanted {}".format(returncode, self._want_return_code)
        )

    def get_params(self) -> Tuple:
        return (self.service_name, self.want_state)

    def describe(self) -> str:
        return "Checking service {} is {}".format(self.service_name, self.want_state)


@register
class MonitorSystemdUnit(Monitor):
    """Monitor a systemd unit.

    This monitor checks the state of the unit as given by
    /org/freedesktop/systemd1/ListUnits
    and reports failure if it is not one of the expected states.
    """

    monitor_type = "systemd-unit"

    # A cached shared by all instances of MonitorSystemdUnit, so a single
    # call is done for all monitors at once.
    _listunit_cache = []  # type: List[Any]
    _listunit_cache_expiry = 0
    CACHE_LIFETIME = 1  # in seconds

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        if not pydbus:
            self.monitor_logger.critical(
                "pydbus package is not available, cannot use MonitorSystemdUnit."
            )
            return
        self.unit_name = cast(str, self.get_config_option("name", required=True))
        self.want_load_states = cast(
            List[str],
            self.get_config_option(
                "load_states", required_type="[str]", default=["loaded"]
            ),
        )
        self.want_active_states = cast(
            List[str],
            self.get_config_option(
                "active_states", required_type="[str]", default=["active", "reloading"]
            ),
        )
        self.want_sub_states = cast(
            List[str],
            self.get_config_option("sub_states", required_type="[str]", default=[]),
        )

    @classmethod
    def _list_units(cls) -> List[Any]:
        if pydbus is None:
            raise RuntimeError("pydbus module not installed")
        if cls._listunit_cache_expiry < time.time():
            bus = pydbus.SystemBus()
            systemd = bus.get(".systemd1")
            cls._listunit_cache_expiry = int(time.time()) + cls.CACHE_LIFETIME
            cls._listunit_cache = list(systemd.ListUnits())
        return cls._listunit_cache

    def run_test(self) -> bool:
        """Check the service is in the desired state."""
        nb_matches = 0
        for unit in self._list_units():
            (
                name,
                desc,
                load_state,
                active_state,
                sub_state,
                follower,
                unit_path,
                job_id,
                job_type,
                job_path,
            ) = unit
            if fnmatch.fnmatch(name, self.unit_name):
                result = self._check_unit(name, load_state, active_state, sub_state)
                nb_matches += 1
                # TODO: is this right?
                return result
        return self.record_fail("No unit %s" % self.unit_name)

    def _check_unit(
        self, name: str, load_state: str, active_state: str, sub_state: str
    ) -> bool:
        if self.want_load_states and load_state not in self.want_load_states:
            return self.record_fail(
                "Unit {0} has load state: {1} (wanted {2})".format(
                    name, load_state, self.want_load_states
                )
            )
        if self.want_active_states and active_state not in self.want_active_states:
            return self.record_fail(
                "Unit {0} has active state: {1} (wanted {2})".format(
                    name, active_state, self.want_active_states
                )
            )
        if self.want_sub_states and sub_state not in self.want_sub_states:
            return self.record_fail(
                "Unit {0} has sub state: {1} (wanted {2})".format(
                    name, sub_state, self.want_sub_states
                )
            )
        # TODO: added since there's no other apparent path to success
        return self.record_success("Implicit success for unit {0}".format(name))

    def get_params(self) -> Tuple:
        return (
            self.unit_name,
            self.want_load_states,
            self.want_active_states,
            self.want_sub_states,
        )

    def describe(self) -> str:
        return "Checks unit %s is running" % self.name


@register
class MonitorProcess(Monitor):
    """Check for a running process."""

    monitor_type = "process"

    def __init__(
        self, name: str = "unnamed", config_options: Optional[dict] = None
    ) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(name=name, config_options=config_options)
        if psutil is None:
            self.monitor_logger.critical("psutil is not installed.")
            self.monitor_logger.critical("Try: pip install -r requirements.txt")
        self.process_name = cast(
            str, self.get_config_option("process_name", required=True)
        )
        self.max_count = cast(
            int, self.get_config_option("max_count", required_type="int", default=-1)
        )
        self.min_count = cast(
            int, self.get_config_option("min_count", required_type="int", default=1)
        )
        self.username = cast(
            Optional[str], self.get_config_option("username", required_type="str")
        )

    @staticmethod
    def _find_process_by_name(
        name: str, username: Optional[str] = None
    ) -> List[psutil.Process]:
        processes = []
        for process in psutil.process_iter(["name", "exe", "cmdline", "username"]):
            if (
                name == process.info["name"]
                or (
                    process.info["exe"]
                    and os.path.basename(process.info["exe"]) == name
                )
                or (process.info["cmdline"] and process.info["cmdline"][0] == name)
            ):
                if username is None or username == process.info["username"]:
                    processes.append(process)
        return processes

    def run_test(self) -> bool:
        if psutil is None:
            return self.record_fail("psutil is not installed")
        processes = self._find_process_by_name(self.process_name, self.username)
        count = len(processes)
        if count == 1:
            message = "1 matching process running"
        else:
            message = "{} matching processes running".format(count)
        if count < self.min_count:
            return self.record_fail(message)
        if self.max_count > -1 and count > self.max_count:
            return self.record_fail(message)
        return self.record_success(message)

    def get_params(self) -> Tuple:
        return (self.process_name, self.min_count, self.max_count, self.username)

    def describe(self) -> str:
        desc = "Checking for at least {} and at most {} processes matching {}".format(
            self.min_count,
            "infinity" if self.max_count == -1 else self.max_count,
            self.process_name,
        )
        if self.username:
            desc = desc + " owned by {}".format(self.username)
        return desc


@register
class MonitorEximQueue(Monitor):
    """Make sure an exim queue isn't too big."""

    monitor_type = "eximqueue"
    max_length = 10
    r = re.compile(r"(?P<count>\d+) matches out of (?P<total>\d+) messages")
    path = "/usr/local/sbin"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.max_length = self.get_config_option(
            "max_length", required_type="int", minimum=1
        )
        path = self.get_config_option("path", default="/usr/local/sbin")
        self.path = os.path.join(path, "exiqgrep")

    def run_test(self) -> bool:
        try:
            _output = subprocess.check_output([self.path, "-xc"])
            output = _output.decode("utf-8")
            for line in output.splitlines():
                matches = self.r.match(line)
                if matches:
                    count = int(matches.group("count"))
                    # total = int(matches.group("total"))
                    if count > self.max_length:
                        if count == 1:
                            return self.record_fail("%d message queued" % count)
                        return self.record_fail("%d messages queued" % count)
                    if count == 1:
                        return self.record_success("%d message queued" % count)
                    return self.record_success("%d messages queued" % count)
            return self.record_fail("Error getting queue size")
        except Exception as error:
            return self.record_fail("Error running exiqgrep: %s" % error)

    def describe(self) -> str:
        return "Checking the exim queue length is < %d" % self.max_length

    def get_params(self) -> Tuple:
        return (self.max_length,)


@register
class MonitorWindowsDHCPScope(Monitor):
    """Checks a Windows DHCP scope to make sure it has sufficient free IPs in the pool."""

    # netsh dhcp server \\SERVER scope SCOPE show clients
    # "No of Clients(version N): N in the Scope

    monitor_type = "dhcpscope"
    max_used = 0
    scope = ""
    server = ""
    r = re.compile(r"No of Clients\(version \d+\): (?P<clients>\d+) in the Scope")

    def __init__(self, name: str, config_options: dict) -> None:
        if not self.is_windows(True):
            raise RuntimeError("DHCPScope monitor requires a Windows platform.")
        super().__init__(name, config_options)
        self.max_used = self.get_config_option(
            "max_used", required_type="int", minimum=1
        )
        self.scope = self.get_config_option("scope", required=True)

    def run_test(self) -> bool:
        try:
            output = str(
                subprocess.check_output(
                    ["netsh", "dhcp", "server", "scope", self.scope, "show", "clients"]
                )
            )
            matches = self.r.search(output)
            if matches:
                clients = int(matches.group("clients"))
                if clients > self.max_used:
                    return self.record_fail("%d clients in scope" % clients)
                return self.record_success("%d clients in scope" % clients)
            return self.record_fail("Error getting client count: no match")
        except Exception as error:
            return self.record_fail("Error getting client count: {0}".format(error))

    def describe(self) -> str:
        return "Checking the DHCP scope has fewer than %d leases" % self.max_used



# type: ignore
import subprocess
import unittest
from unittest.mock import MagicMock, patch

from simplemonitor.Monitors import gmirror

DEFAULT_CONFIG_OPTIONS = {"array_device": "gm0", "expected_disks": 2}

MOCK_OUTPUT_GOOD = """gm0  COMPLETE  ada0 (ACTIVE)
gm0  COMPLETE  ada1 (ACTIVE)
"""
MOCK_OUTPUT_SYNCHRONIZING = """gm0  DEGRADED  ada0 (ACTIVE)
gm0  DEGRADED  ada1 (SYNCHRONIZING, 7%)
"""
MOCK_OUTPUT_BAD = """gm0  DEGRADED  ada0 (ACTIVE)
"""


class TestGmirrorStatusMonitors(unittest.TestCase):
    @patch("subprocess.run")
    def test_GmirrorStatus_success(self, mock_run):
        """Success / happy path tests."""

        mock_stdout = MagicMock()
        mock_stdout.configure_mock(**{"stdout.decode.return_value": MOCK_OUTPUT_GOOD})
        mock_run.return_value = mock_stdout

        m = gmirror.MonitorGmirrorStatus("test", DEFAULT_CONFIG_OPTIONS)

        m.run_test()

        mock_run.assert_called_with(
            ["gmirror", "status", "-gs", DEFAULT_CONFIG_OPTIONS.get("array_device")],
            capture_output=True,
            check=True,
        )
        self.assertEqual("Array gm0 is in state COMPLETE with 2 disks", m.get_result())
        self.assertTrue(m.test_success())
        self.assertEqual(m.error_count, 0)

        m.run_test()

        self.assertTrue(m.test_success())
        self.assertEqual(m.error_count, 0)

    @patch("subprocess.run")
    def test_GmirrorStatus_failedSynchronizing(self, mock_run):
        """Check failure test cases."""

        mock_stdout = MagicMock()
        mock_stdout.configure_mock(
            **{"stdout.decode.return_value": MOCK_OUTPUT_SYNCHRONIZING}
        )
        mock_run.return_value = mock_stdout

        m = gmirror.MonitorGmirrorStatus("test", DEFAULT_CONFIG_OPTIONS)

        m.run_test()

        mock_run.assert_called()
        self.assertEqual(m.get_result(), "Array gm0 is in state DEGRADED with 2 disks")
        self.assertFalse(m.test_success())
        self.assertEqual(m.error_count, 1)

        m.run_test()

        self.assertFalse(m.test_success())
        self.assertEqual(m.error_count, 2)

    @patch("subprocess.run")
    def test_GmirrorStatus_failedMissingDisk(self, mock_run):
        """Check failure test cases."""

        mock_stdout = MagicMock()
        mock_stdout.configure_mock(**{"stdout.decode.return_value": MOCK_OUTPUT_BAD})
        mock_run.return_value = mock_stdout

        m = gmirror.MonitorGmirrorStatus("test", DEFAULT_CONFIG_OPTIONS)

        m.run_test()

        mock_run.assert_called()
        self.assertEqual(m.get_result(), "Array gm0 is in state DEGRADED with 1 disks")
        self.assertFalse(m.test_success())
        self.assertEqual(m.error_count, 1)

    @patch("subprocess.run")
    def test_GmirrorStatus_raiseFileNotFoundError(self, mock_run):
        """Make sure the program raises if the binary isn't present at all."""

        mock_run.side_effect = FileNotFoundError(
            "[Errno 2] No such file or directory: 'gmirror'"
        )
        m = gmirror.MonitorGmirrorStatus("test", DEFAULT_CONFIG_OPTIONS)
        self.assertRaises(FileNotFoundError, m.run_test)

    @patch("subprocess.run")
    def test_GmirrorStatus_failureSubprocessFailure(self, mock_run):
        """Handle failure based on command failing with non-0 exit code."""

        mock_run.side_effect = subprocess.CalledProcessError(
            1,
            ["gmirror", "status", "-gs", DEFAULT_CONFIG_OPTIONS.get("array_device")],
        )
        m = gmirror.MonitorGmirrorStatus("test", DEFAULT_CONFIG_OPTIONS)
        m.run_test()
        self.assertFalse(m.test_success())
        self.assertEqual(m.error_count, 1)


if __name__ == "__main__":
    unittest.main()



"""
SimpleMonitor alerts via SMS77
"""

from typing import cast

import requests

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class SMS77Alerter(Alerter):
    """Send SMS alerts using the sms77 service"""

    alerter_type = "sms77"
    urgent = True

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.api_key = cast(
            str, self.get_config_option("api_key", required=True, allow_empty=False)
        )
        self.target = cast(
            str, self.get_config_option("target", required=True, allow_empty=False)
        )

        self.sender = cast(str, self.get_config_option("sender", default="SmplMntr"))
        if len(self.sender) > 11:
            self.alerter_logger.warning("truncating SMS sender name to 11 chars")
            self.sender = self.sender[:11]

        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

        self.support_catchup = True

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Send an SMS alert"""

        alert_type = self.should_alert(monitor)
        if alert_type not in [AlertType.FAILURE, AlertType.SUCCESS]:
            return

        message = self.build_message(AlertLength.SMS, alert_type, monitor)

        url = "https://gateway.sms77.io/api/sms"
        params = {
            "text": message,
            "to": self.target,
            "from": self.sender,
            "p": self.api_key,
        }

        if not self._dry_run:
            try:
                response = requests.get(url, params=params, timeout=self.timeout)
                status = response.text
                if not status.startswith("100"):
                    self.alerter_logger.error(
                        "Unable to send SMS: status code %s", status
                    )
            except requests.RequestException:
                self.alerter_logger.exception("SMS sending failed")
        else:
            self.alerter_logger.info("dry_run: would send SMS: %s", message)

    def _describe_action(self) -> str:
        return "SMSing {target} via SMS77".format(target=self.target)



"""
SimpleMonitor alerts via healthchecks
"""

import json
from typing import cast

import requests

from ..Monitors.monitor import Monitor
from ..util import check_group_match
from .alerter import Alerter, AlertLength, AlertType, register


@register
class HealthchecksAlerter(Alerter):
    """Send push notification via Healthchecks."""

    alerter_type = "healthchecks"
    headers = {"User-Agent": "SimpleMonitor"}

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.token = cast(
            str, self.get_config_option("token", required=True, allow_empty=False)
        )
        self.create = cast(str, self.get_config_option("create", required_type="bool"))
        hc_headers = cast(str, self.get_config_option("headers"))
        if hc_headers:
            try:
                self.headers = json.loads(hc_headers)
            except json.JSONDecodeError as e:
                self.alerter_logger.error(f"Parsing headers to JSON failed: {e}")

        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

        self.support_catchup = True

    def send_notification(
        self,
        subject: str,
        body: str,
        alert_type: str,
        name: str,
        slug: str,
        dry_run=False,
    ) -> None:
        """Send a push notification."""
        url = "https://hc-ping.com/" + self.token + "/" + slug
        if alert_type != "success":
            url += "/fail"
        if self.create:
            url += "?create=1"
        if dry_run:
            self.alerter_logger.info(
                f"dry_run - monitor: {self.name} | alerter: {name} | {url}"
            )
            return

        data = ""
        data_json = {}
        if subject:
            data_json["subject"] = subject
        if body:
            data = body
            data_json["body"] = body.strip()

        try:
            req = requests.post(
                url,
                data=data,
                # json=data_json,
                headers=self.headers,
                timeout=self.timeout,
            )
            req.raise_for_status()
            self.alerter_logger.info(f"monitor: {name} | alerter: {self.name} | {url}")
        except requests.exceptions.RequestException:
            raise ValueError(
                f"alerter: {self.name} | {slug} | {req.status_code}"
            ) from None

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Build up the content for the push notification."""

        if monitor.slug is None or not monitor.enabled:
            return

        if not check_group_match(monitor.group, self.groups):
            return

        alert_type = self.should_alert(monitor)

        if alert_type == AlertType.NONE:
            if monitor.failures:
                alert_type = AlertType.CATCHUP
            else:
                alert_type = AlertType.SUCCESS

        subject = self.build_message(AlertLength.NOTIFICATION, alert_type, monitor)
        body = self.build_message(AlertLength.FULL, alert_type, monitor)

        if not self._dry_run:
            try:
                self.send_notification(
                    subject, body, alert_type.value, monitor.name, monitor.slug
                )
            except Exception as e:
                self.alerter_logger.error(
                    f"Notification failed: monitor: {monitor.name} | group: {monitor.group} | {e}"
                )
        else:
            self.send_notification(
                subject,
                body,
                alert_type.value,
                monitor.name,
                monitor.slug,
                dry_run=self._dry_run,
            )

    def _describe_action(self) -> str:
        return "posting to healthchecks"



"""
SimpleMonitor alerts via pushover
"""

from typing import cast

import requests

from ..Monitors.monitor import Monitor
from .alerter import Alerter, AlertLength, AlertType, register


@register
class PushoverAlerter(Alerter):
    """Send push notification via Pushover."""

    alerter_type = "pushover"

    def __init__(self, config_options: dict) -> None:
        super().__init__(config_options)
        self.pushover_token = cast(
            str, self.get_config_option("token", required=True, allow_empty=False)
        )
        self.pushover_user = cast(
            str, self.get_config_option("user", required=True, allow_empty=False)
        )
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

        self.support_catchup = True

    def send_pushover_notification(self, subject: str, body: str) -> None:
        """Send a push notification."""
        requests.post(
            "https://api.pushover.net/1/messages.json",
            data={
                "token": self.pushover_token,
                "user": self.pushover_user,
                "title": subject,
                "message": body,
            },
            timeout=self.timeout,
        )

    def send_alert(self, name: str, monitor: Monitor) -> None:
        """Build up the content for the push notification."""

        alert_type = self.should_alert(monitor)

        if alert_type == AlertType.NONE:
            return
        subject = self.build_message(AlertLength.NOTIFICATION, alert_type, monitor)
        body = self.build_message(AlertLength.FULL, alert_type, monitor)

        if not self._dry_run:
            try:
                self.send_pushover_notification(subject, body)
            except Exception:
                self.alerter_logger.exception("Couldn't send push notification")
        else:
            self.alerter_logger.info("dry_run: would send push notification: %s", body)

    def _describe_action(self) -> str:
        return "posting to pushover"



# type: ignore
import datetime
import os
import textwrap
import unittest
from unittest import mock

from freezegun import freeze_time

from simplemonitor import util
from simplemonitor.Alerters import alerter, sns
from simplemonitor.Monitors import monitor

# Create a consistent "local" timezone and offset for the tests, for tests that
# compare the offset between UTC and local time. For simplicity and
# predictability, use a time zone that doesn't have daylight savings.
TZ_LOCAL = "America/Phoenix"
TZ_LOCAL_OFFSET = -7
TZ_UTC = "UTC"


# Default to UTC, then override for tests where we need to have timezone
# specific logic.
@mock.patch.dict(os.environ, {"TZ": TZ_UTC}, clear=True)
class TestAlerter(unittest.TestCase):
    def test_groups(self):
        config_options = {"groups": "a,b,c"}
        a = alerter.Alerter(config_options)
        self.assertEqual(["a", "b", "c"], a.groups)

    def test_times_always(self):
        config_options = {"times_type": "always"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._times_type, alerter.AlertTimeFilter.ALWAYS)
        self.assertEqual(a._time_info, (None, None))

    def test_times_only(self):
        config_options = {
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config_options)
        self.assertEqual(a._times_type, alerter.AlertTimeFilter.ONLY)
        self.assertEqual(a._time_info, (datetime.time(10, 00), datetime.time(11, 00)))

    def test_times_not(self):
        config_options = {
            "times_type": "not",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config_options)
        self.assertEqual(a._times_type, alerter.AlertTimeFilter.NOT)
        self.assertEqual(a._time_info, (datetime.time(10, 00), datetime.time(11, 00)))

    def test_times_broken(self):
        config_options = {
            "times_type": "fake",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        with self.assertRaises(ValueError):
            alerter.Alerter(config_options)

    def test_days(self):
        config_options = {"days": "0,1,4"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._days, [0, 1, 4])

    def test_delay(self):
        config_options = {"delay": "1"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._delay_notification, True)

    def test_dryrun(self):
        config_options = {"dry_run": "1"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._dry_run, True)

    def test_oohrecovery(self):
        config_otions = {"ooh_recovery": "1"}
        a = alerter.Alerter(config_otions)
        self.assertEqual(a._ooh_recovery, True)

    def test_dependencies(self):
        config_options = {"depend": "a,b,c"}
        a = alerter.Alerter(config_options)
        self.assertEqual(
            a.dependencies, ["a", "b", "c"], "Alerter did not store dependencies"
        )
        self.assertEqual(
            a.check_dependencies(["d", "e"]), True, "Alerter thinks a dependency failed"
        )
        self.assertEqual(
            a.check_dependencies(["a"]),
            False,
            "Alerter did not notice a dependency failed",
        )

    def test_limit(self):
        config_options = {"limit": "5"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._limit, 5)

    def test_repeat(self):
        config_options = {"repeat": "5"}
        a = alerter.Alerter(config_options)
        self.assertEqual(a._repeat, 5)

    def test_should_alert_basic_failure(self):
        # no special alert config
        a = alerter.Alerter(None)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_should_alert_only_failure(self):
        # no special alert config
        a = alerter.Alerter({"only_failures": True})
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_should_alert_basic_none(self):
        a = alerter.Alerter(None)
        m = monitor.MonitorNull()
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    def test_should_alert_basic_success(self):
        a = alerter.Alerter(None)
        m = monitor.MonitorFail("fail", {})
        for _ in range(0, 6):
            m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.SUCCESS)

    def test_should_not_alert_basic_success(self):
        a = alerter.Alerter({"only_failures": True})
        m = monitor.MonitorFail("fail", {})
        for _ in range(0, 6):
            m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    @freeze_time("2020-03-10 22:00")  # A Tuesday
    def test_not_allowed_today(self):
        a = alerter.Alerter({"days": "0,2,3,4,5,6"})
        self.assertFalse(a._allowed_today())

    @freeze_time("2020-03-09 20:00")  # Monday UTC, but TZ will push to Tuesday
    def test_not_allowed_today_tz(self):
        """
        Test that we handle timezone conversion properly with disallowed days.
        """
        # Note: This doesn't map with either the regular "local" timezone _or_
        # GMT; that's why we're explicitly setting times_tz offset to a
        # positive value.
        a = alerter.Alerter({"days": "0,2,3,4,5,6", "times_tz": "+05:00"})
        self.assertFalse(a._allowed_today())

    @freeze_time("2020-03-10 22:00")
    def test_allowed_today(self):
        """
        Test that we handle timezone conversion properly with allowed days.
        """
        a = alerter.Alerter({"days": "1"})
        self.assertTrue(a._allowed_today())

    @freeze_time("2020-03-09 20:00")  # Monday UTC, but TZ will push to Tuesday
    def test_allowed_today_tz(self):
        a = alerter.Alerter({"days": "1", "times_tz": "+05:00"})
        self.assertTrue(a._allowed_today())

    @freeze_time("2020-03-10")
    def test_allowed_default(self):
        a = alerter.Alerter({})
        self.assertTrue(a._allowed_today())

    @freeze_time("10:00")
    def test_allowed_always(self):
        a = alerter.Alerter({})
        self.assertTrue(a._allowed_time())

    def test_allowed_only(self):
        """
        Test logic with the "only" schedule (only between lower and upper
        bounds).
        """
        a = alerter.Alerter(
            {
                "times_type": "only",
                "time_lower": "10:00",
                "time_upper": "11:00",
            }
        )
        with freeze_time("09:00"):
            self.assertFalse(a._allowed_time())
        with freeze_time("10:30"):
            self.assertTrue(a._allowed_time())
        with freeze_time("12:00"):
            self.assertFalse(a._allowed_time())

    # Influence time_tz indirectly, though we could also set it directly in
    # alerter.Alerter() below.
    @mock.patch.dict(os.environ, {"TZ": TZ_LOCAL}, clear=True)
    def test_allowed_only_tz(self):
        """Test `times_type=only` with the time in local time."""
        a = alerter.Alerter(
            {
                "times_type": "only",
                "time_lower": "09:00",  # 9:00 America/Phoenix, 16:00 UTC
                "time_upper": "10:00",  # 10:00 America/Phoenix, 17:00 UTC
            }
        )
        with freeze_time("15:00"):
            self.assertFalse(a._allowed_time())
        with freeze_time("16:30"):
            self.assertTrue(a._allowed_time())
        with freeze_time("18:00"):
            self.assertFalse(a._allowed_time())

    def test_allowed_not(self):
        """Test when using `times_type=not`."""
        a = alerter.Alerter(
            {
                "times_type": "not",
                "time_lower": "10:00",
                "time_upper": "11:00",
            }
        )
        with freeze_time("09:00"):
            self.assertTrue(a._allowed_time())
        with freeze_time("10:30"):
            self.assertFalse(a._allowed_time())
        with freeze_time("12:00"):
            self.assertTrue(a._allowed_time())

    @mock.patch.dict(os.environ, {"TZ": TZ_LOCAL}, clear=True)
    def test_allowed_not_tz(self):
        """Test this variant for a specific timezone."""
        a = alerter.Alerter(
            {
                "times_type": "not",
                "time_lower": "09:00",  # 9:00 America/Phoenix, 16:00 UTC
                "time_upper": "10:00",  # 10:00 America/Phoenix, 17:00 UTC
            }
        )
        with freeze_time("15:55"):
            self.assertTrue(a._allowed_time())
        with freeze_time("16:30"):
            self.assertFalse(a._allowed_time())
        with freeze_time("17:00"):
            self.assertTrue(a._allowed_time())

    def test_should_not_alert_ooh(self):
        """Make sure we don't alert outside of hours."""
        config = {
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        with freeze_time("2020-03-10 09:00"):
            # Out of hours on the right day; shouldn't alert
            self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
            self.assertEqual(a._ooh_failures, ["fail"])

        a = alerter.Alerter(config)
        with freeze_time("2020-03-10 12:00"):
            self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
            self.assertEqual(a._ooh_failures, ["fail"])

    @freeze_time("2020-03-10 10:30")
    def test_should_alert_ooh(self):
        """Make sure we do alert within scheduled hours."""
        config = {
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)
        self.assertEqual(a._ooh_failures, [])

    @freeze_time("2020-03-10 10:30")
    def test_should_alert_limit(self):
        config = {
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
            "limit": 2,
        }
        a = alerter.Alerter(config)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
        self.assertEqual(a._ooh_failures, [])

        # It should only alert on the $limit attempt
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)
        self.assertEqual(a._ooh_failures, [])

        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
        self.assertEqual(a._ooh_failures, [])

    @freeze_time("2020-03-10 09:00")
    def test_should_alert_limit_ooh(self):
        config = {
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
            "limit": 2,
        }
        a = alerter.Alerter(config)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
        self.assertEqual(a._ooh_failures, [])

        a = alerter.Alerter(config)
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
        self.assertEqual(a._ooh_failures, ["fail"])

        a = alerter.Alerter(config)
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
        self.assertEqual(a._ooh_failures, [])

    def test_should_alert_catchup(self):
        """
        Test that, with `delay` 1 and `support_catchup` = `True`, an alert will
        be sent if a failure from before the period in which notifications are
        allowed, continues once we're within the notification window.
        """
        config = {
            "delay": 1,
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config)
        a.support_catchup = True
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        with freeze_time("2020-03-10 09:00"):
            self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
            self.assertEqual(a._ooh_failures, ["fail"])
        with freeze_time("2020-03-10 10:30"):
            self.assertEqual(a.should_alert(m), alerter.AlertType.CATCHUP)
            self.assertEqual(a._ooh_failures, [])

    def test_should_alert_no_catchup(self):
        """
        Here, `support_catchup` is unset, so catchup notifications shouldn't be
        sent.
        """
        config = {
            "delay": 1,
            "times_type": "only",
            "time_lower": "10:00",
            "time_upper": "11:00",
        }
        a = alerter.Alerter(config)
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        with freeze_time("2020-03-10 09:00"):
            self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)
            self.assertEqual(a._ooh_failures, ["fail"])
        with freeze_time("2020-03-10 10:30"):
            self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)
            self.assertEqual(a._ooh_failures, [])

    def test_skip_group_alerter(self):
        a = alerter.Alerter({"groups": "test"})
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    def test_skip_group_monitor(self):
        a = alerter.Alerter()
        m = monitor.MonitorFail("fail", {"group": "test"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    def test_groups_match(self):
        a = alerter.Alerter({"groups": "test"})
        m = monitor.MonitorFail("fail", {"group": "test"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_groups_multi_match(self):
        a = alerter.Alerter({"groups": "test1, test2"})
        m = monitor.MonitorFail("fail", {"group": "test1"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_groups_all_match(self):
        a = alerter.Alerter({"groups": "_all"})
        m = monitor.MonitorFail("fail", {"group": "test1"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_disabled_monitor_no_alert(self):
        a = alerter.Alerter()
        m = monitor.MonitorFail("fail", {"enabled": 0})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    def test_alert_not_urgent(self):
        a = alerter.Alerter()
        m = monitor.MonitorFail("fail", {})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_no_alert_urgent(self):
        a = alerter.Alerter({"urgent": "1"})
        m = monitor.MonitorFail("fail", {"urgent": "0"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.NONE)

    def test_alert_urgent(self):
        a = alerter.Alerter({"urgent": "1"})
        m = monitor.MonitorFail("fail", {"urgent": "1"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)

    def test_alert_alerter_not_urgent(self):
        a = alerter.Alerter({"urgent": "0"})
        m = monitor.MonitorFail("fail", {"urgent": "1"})
        m.run_test()
        self.assertEqual(a.should_alert(m), alerter.AlertType.FAILURE)


class TestMessageBuilding(unittest.TestCase):
    def setUp(self):
        # Because we're instantiating the `alerter.Alerter()` object here, and
        # because the decorator isn't consumed in the setUp routine, need to
        # use `with` here instead for patching the time zone.
        with mock.patch.dict(os.environ, {"TZ": TZ_UTC}, clear=True):
            self.test_alerter = alerter.Alerter()
            self.expected_time_string = "2020-03-10 09:00:00+00:00"

    def test_notification_format_failure(self):
        m = monitor.MonitorFail("test", {})
        m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.NOTIFICATION, alerter.AlertType.FAILURE, m
            ),
            "Monitor test on %s failed" % self.test_alerter.hostname,
        )

    def test_notification_format_success(self):
        m = monitor.MonitorNull("winning", {})
        for _ in range(0, 6):
            m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.NOTIFICATION, alerter.AlertType.SUCCESS, m
            ),
            "Monitor winning on %s succeeded" % self.test_alerter.hostname,
        )

    @freeze_time("2020-03-10 09:00")
    def test_oneline_format_failure(self):
        m = monitor.MonitorFail("test", {})
        m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.ONELINE, alerter.AlertType.FAILURE, m
            ),
            "failure: test on {hostname} failed at {expected_time} (0+00:00:00): This monitor always fails.".format(
                hostname=util.short_hostname(),
                expected_time=self.expected_time_string,
            ),
        )

    def test_oneline_format_success(self):
        m = monitor.MonitorNull("winning", {})
        for _ in range(0, 6):
            m.run_test()
        m.last_result = "a " * 80
        desired = (
            "success: winning on {hostname} succeeded at  (0+00:00:00): ".format(
                hostname=util.short_hostname()
            )
            + "a " * 80
        )
        output = self.test_alerter.build_message(
            alerter.AlertLength.ONELINE, alerter.AlertType.SUCCESS, m
        )
        self.assertEqual(desired, output)

    @freeze_time("2020-03-10 09:00")
    def test_sms_format_failure(self):
        m = monitor.MonitorFail("test", {})
        m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.SMS, alerter.AlertType.FAILURE, m
            ),
            "failure: test on {hostname} failed at {expected_time} (0+00:00:00): This monitor always fails.".format(
                hostname=util.short_hostname(),
                expected_time=self.expected_time_string,
            ),
        )

    def test_sms_format_success(self):
        m = monitor.MonitorNull("winning", {})
        for _ in range(0, 6):
            m.run_test()
        m.last_result = "a " * 80
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.SMS, alerter.AlertType.SUCCESS, m
            ),
            textwrap.shorten(
                "success: winning on {hostname} succeeded at (0+00:00:00): {a}".format(
                    hostname=util.short_hostname(), a=m.last_result
                ),
                width=160,
                placeholder="...",
            ),
        )

    @freeze_time("2020-03-10 09:00")
    def test_full_format_failure(self):
        m = monitor.MonitorFail("test", {})
        m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.FULL, alerter.AlertType.FAILURE, m
            ),
            textwrap.dedent(
                """\
                Monitor test on {hostname} failed!
                Failed at: {expected_time} (down 0+00:00:00)
                Virtual failure count: 1
                Additional info: This monitor always fails.
                Description: A monitor which always fails.
                """.format(
                    expected_time=self.expected_time_string,
                    hostname=self.test_alerter.hostname,
                )
            ),
        )

    @freeze_time("2020-03-10 09:00")
    def test_full_format_failure_docs(self):
        m = monitor.MonitorFail("test", {"failure_doc": "whoops"})
        m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.FULL, alerter.AlertType.FAILURE, m
            ),
            textwrap.dedent(
                """\
                Monitor test on {host} failed!
                Failed at: {expected_time} (down 0+00:00:00)
                Virtual failure count: 1
                Additional info: This monitor always fails.
                Description: A monitor which always fails.
                Documentation: whoops
                """.format(
                    expected_time=self.expected_time_string,
                    host=self.test_alerter.hostname,
                )
            ),
        )

    @freeze_time("2020-03-10 09:00")
    def test_full_format_success(self):
        m = monitor.MonitorNull("winning", {})
        for _ in range(0, 6):
            m.run_test()
        self.assertEqual(
            self.test_alerter.build_message(
                alerter.AlertLength.FULL, alerter.AlertType.SUCCESS, m
            ),
            textwrap.dedent(
                """\
                Monitor winning on {host} succeeded!
                Recovered at: {expected_time} (was down for 0+00:00:00)
                Additional info: 
                Description: (Monitor did not write an auto-biography.)
                """.format(  # noqa: W291
                    expected_time=self.expected_time_string,
                    host=self.test_alerter.hostname,
                )
            ),
        )

    def test_was_downtime(self):
        m = monitor.MonitorFail("test", {})
        with freeze_time("2020-03-10 09:00") as frozen_time:
            for _ in range(0, 6):
                m.run_test()
                frozen_time.tick(30)
            self.assertEqual(m.get_wasdowntime(), util.UpDownTime(0, 0, 2, 30))


class TestMessageBuildingTZ(TestMessageBuilding):
    def setUp(self):
        super().setUp()
        self.test_alerter._tz = "Europe/Warsaw"
        self.expected_time_string = "2020-03-10 10:00:00+01:00"


class TestSNSAlerter(unittest.TestCase):
    def test_config(self):
        with self.assertRaises(util.AlerterConfigurationError):
            sns.SNSAlerter({})
        with self.assertRaises(util.AlerterConfigurationError):
            sns.SNSAlerter({"topic": "a", "number": "b"})

    def test_urgent(self):
        a = sns.SNSAlerter({"topic": "a"})
        self.assertEqual(a.urgent, True)

    def test_not_urgent(self):
        a = sns.SNSAlerter({"topic": "a", "urgent": 0})
        self.assertEqual(a.urgent, False)


@mock.patch.dict(os.environ, {"TZ": TZ_LOCAL}, clear=True)
class TestDescription(unittest.TestCase):
    def test_times_always(self):
        a = alerter.Alerter({"times_type": "always"})
        self.assertEqual(a._describe_times(), "(always)")

    def test_times_only_times(self):
        a = alerter.Alerter(
            {"times_type": "only", "time_lower": "09:00", "time_upper": "10:00"}
        )
        self.assertEqual(
            a._describe_times(), f"only between 09:00 and 10:00 ({TZ_LOCAL}) on any day"
        )

    def test_times_only_days(self):
        a = alerter.Alerter(
            {
                "times_type": "only",
                "time_lower": "09:00",
                "time_upper": "10:00",
                "days": "0,1,2",
            }
        )
        self.assertEqual(
            a._describe_times(),
            f"only between 09:00 and 10:00 ({TZ_LOCAL}) on Mon, Tue, Wed",
        )

    def test_times_not_time(self):
        a = alerter.Alerter(
            {"times_type": "not", "time_lower": "09:00", "time_upper": "10:00"}
        )
        self.assertEqual(
            a._describe_times(),
            f"any time except between 09:00 and 10:00 ({TZ_LOCAL}) on any day",
        )

    def test_times_not_days(self):
        a = alerter.Alerter(
            {
                "times_type": "not",
                "time_lower": "09:00",
                "time_upper": "10:00",
                "days": "3,4,5,6",
            }
        )
        self.assertEqual(
            a._describe_times(),
            f"any time except between 09:00 and 10:00 ({TZ_LOCAL}) on Thu, Fri, Sat, Sun",
        )



"""
Home Automation monitors for SimpleMonitor
"""

from typing import Tuple, cast

import requests

from .monitor import Monitor, register


@register
class MonitorSensor(Monitor):
    """Monitor the existence of a HASS sensor"""

    monitor_type = "hass_sensor"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.url = cast(str, self.get_config_option("url", required=True))
        self.sensor = cast(str, self.get_config_option("sensor", required=True))
        self.token = cast(str, self.get_config_option("token", default=None))
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )

    def describe(self) -> str:
        return "monitor the existence of a sensor"

    def run_test(self) -> bool:
        try:
            # retrieve the status from hass API
            self.monitor_logger.debug(
                requests.get(
                    f"{self.url}/api/states/{self.sensor}", timeout=self.timeout
                ).text
            )
            call = requests.get(
                f"{self.url}/api/states/{self.sensor}",
                timeout=self.timeout,
                headers={
                    "Authorization": f"Bearer {self.token}",
                    "Content-Type": "application/json",
                },
            )
            if not call.ok:
                raise ValueError(call.text)
            response = call.json()
            self.monitor_logger.debug("retrieved JSON: %s", response)
        except requests.RequestException as error:
            # a general issue getting to the API
            # nothing special to report, this monitor should be configured to be
            # dependent of general hass API availability
            return self.record_fail(f"cannot get info from hass: {error}")
        else:
            # we have a response from the API
            # now: is the sensor defined at all in hass? If not the answer is basically empty
            if response.get("context"):
                if response["state"] == "unavailable":
                    return self.record_fail(
                        "the sensor exists but state is 'unavailable'"
                    )
                return self.record_success()
            return self.record_fail("sensor not found in hass")

    def get_params(self) -> Tuple:
        return (self.url, self.sensor)



"""
Simplemonitor logger for seq

Inspiration from
https://raw.githubusercontent.com/eifinger/appdaemon-scripts/master/seqSink/seqSink.py
"""

import datetime
import json
from typing import Optional, cast

import requests

from ..Monitors.monitor import Monitor
from .logger import Logger, register


@register
class SeqLogger(Logger):
    """Logging to seq"""

    logger_type = "seq"
    only_failures = False
    buffered = False
    dateformat = None

    def __init__(self, config_options: Optional[dict] = None) -> None:
        if config_options is None:
            config_options = {}
        super().__init__(config_options)

        # i.e. http://192.168.0.5:5341
        self.endpoint = cast(
            str, self.get_config_option("endpoint", required=True, allow_empty=False)
        )
        self.timeout = cast(
            int, self.get_config_option("timeout", required_type="int", default=5)
        )
        # Potentially, would need to add a header for ApiKey

        # Send message to indicate we have started logging
        self.log_to_seq(
            self.endpoint,
            "SeqLogger",
            "simpleMonitor",
            "__init__",
            None,
            "logging enabled for simpleMonitor",
            False,
        )

    def save_result2(self, name: str, monitor: Monitor) -> None:
        try:
            is_fail = monitor.test_success() is False

            self.log_to_seq(
                self.endpoint,
                name,
                monitor.name,
                monitor.monitor_type,
                str(monitor.get_params()),
                monitor.describe(),
                is_fail,
            )
        except Exception:
            self.logger_logger.exception("Error sending to seq in %s", monitor.name)

    def describe(self) -> str:
        return "Sends simple log to seq using raw endpoint"

    def log_to_seq(
        self, endpoint, name, app_name, monitor_type, params, description, is_fail
    ):
        """Send an event to seq"""
        event_data = {
            "Timestamp": str(datetime.datetime.now()),
            "Level": "Error" if is_fail is True else "Information",
            "MessageTemplate": str(description),
            "Properties": {
                "Type": "simpleMonitor",
                "Name": name,
                "Monitor": str(app_name),
                "MonitorType": monitor_type,
                # "Params": params
            },
        }
        if params is not None:
            event_data["Properties"]["Params"] = params

        request_body = {"Events": [event_data]}

        try:
            _ = json.dumps(request_body)  # This just checks it is valid...
        except TypeError:
            self.logger_logger.error("Could not serialise %s", request_body)
            return

        try:
            response = requests.post(
                self.endpoint, json=request_body, timeout=self.timeout
            )
            if not response.status_code == 200 and not response.status_code == 201:
                self.logger_logger.error(
                    "POST to seq failed with status code: %s", response
                )
        except requests.RequestException:
            self.logger_logger.exception("Failed to log to seq")



"""A version of ConfigParser which supports subsitutions from environment variables."""

import os
import re
from configparser import BasicInterpolation, ConfigParser
from typing import Any, List, Optional


class EnvironmentAwareConfigParser(ConfigParser):
    """A subclass of ConfigParser which allows %env:VAR% interpolation via the
    get method."""

    r = re.compile("%env:([a-zA-Z0-9_]+)%")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Init with our specific interpolation class (for Python 3)"""
        interpolation = EnvironmentAwareInterpolation()
        kwargs["interpolation"] = interpolation
        ConfigParser.__init__(self, *args, **kwargs)

    def read(self, filenames: Any, encoding: Optional[str] = None) -> List[str]:
        """Load a config file and do environment variable interpolation on the section names."""
        result = ConfigParser.read(self, filenames)
        for section in self.sections():
            original_section = section
            matches = self.r.search(section)
            while matches:
                env_key = matches.group(1)
                if env_key in os.environ:
                    section = section.replace(matches.group(0), os.environ[env_key])
                else:
                    raise ValueError(
                        "Cannot find {0} in environment for config interpolation".format(
                            env_key
                        )
                    )

                matches = self.r.search(section)
            if section != original_section:
                self.add_section(section)
                for option, value in self.items(original_section):
                    self.set(section, option, value)
                self.remove_section(original_section)
        return result


class EnvironmentAwareInterpolation(BasicInterpolation):
    """An interpolation which substitutes values from the environment."""

    r = re.compile("%env:([a-zA-Z0-9_]+)%")

    def before_get(
        self, parser: Any, section: str, option: str, value: Any, defaults: Any
    ) -> Any:
        parser.get(section, option, raw=True, fallback=value)
        matches = self.r.search(value)
        old_value = value
        while matches:
            env_key = matches.group(1)
            if env_key in os.environ:
                value = value.replace(matches.group(0), os.environ[env_key])
            else:
                raise ValueError(
                    "Cannot find {0} in environment for config interpolation".format(
                        env_key
                    )
                )
            matches = self.r.search(value)
            if value == old_value:
                break
            old_value = value
        return value



"""
Ring doorbell battery monitoring for SimpleMonitor
"""

# mypy: ignore-errors
# ^-- remove this once the ring-doorbell library issue is fixed

import json
from pathlib import Path
from typing import Optional, cast

import ring_doorbell
from oauthlib.oauth2.rfc6749.errors import MissingTokenError

from ..Monitors.monitor import Monitor, register
from ..version import VERSION

RING_USER_AGENT = "SimpleMonitor/{}".format(VERSION)


@register
class MonitorRingDoorbell(Monitor):
    """Monitor the battery life on a Ring doorbell."""

    monitor_type = "ring_doorbell"

    def __init__(self, name: str, config_options: dict) -> None:
        raise RuntimeError(
            "RingDoorbell needs fixing; see https://github.com/jamesoff/simplemonitor/issues/1624"
        )
        if "gap" not in config_options:
            config_options["gap"] = 21600  # 6 hours
        super().__init__(name, config_options)
        self.device_name = cast(str, self.get_config_option("device_name"))
        self.device_type = cast(
            str, self.get_config_option("device_type", default="doorbell")
        )
        self.minimum_battery = cast(
            int,
            self.get_config_option("minimum_battery", required_type="int", default=25),
        )
        self.ring_username = cast(
            str, self.get_config_option("username", required=True)
        )
        self.ring_password = cast(
            str, self.get_config_option("password", required=True)
        )
        self.cache_file = Path(
            self.get_config_option("cache_file", default=".ring_token.cache")
        )
        if self.cache_file.is_file():
            self.monitor_logger.info("Using token cache file for Ring")
            self._ring_auth = ring_doorbell.Auth(
                RING_USER_AGENT,
                json.loads(self.cache_file.read_text()),
                self._token_updated,
            )
        else:
            self._ring_auth = ring_doorbell.Auth(
                RING_USER_AGENT, token_updater=self._token_updated
            )
            try:
                self.monitor_logger.info("Logging in to Ring")
                self._ring_auth.fetch_token(self.ring_username, self.ring_password)
            except MissingTokenError:
                self.monitor_logger.critical("MFA logins are not supported")
                self._ring_auth = None
        self.ring = None  # type: Optional[ring_doorbell.Ring]

    def _token_updated(self, token: str) -> None:
        self.cache_file.write_text(json.dumps(token))

    def run_test(self) -> bool:
        if self.ring is None:
            self.ring = ring_doorbell.Ring(self._ring_auth)
        self.ring.update_data()
        devices = self.ring.devices()
        # doorbots are doorbells owned by this account
        # authorized_doorbots are ones shared with this account
        # the device of interest could be in either depending on how the API
        # user we're configured with relates to it
        if self.device_type == "doorbell":
            doorbells = devices["authorized_doorbots"]
            doorbells.extend(devices["doorbots"])
        elif self.device_type == "camera":
            doorbells = devices.get("authorized_stickup_cams", [])
            doorbells.extend(devices["stickup_cams"])
        else:
            return self.record_fail(f"Unknown device type {self.device_type}")
        for doorbell in doorbells:
            if doorbell.name == self.device_name:
                battery = int(doorbell.battery_life)
                if battery < self.minimum_battery:
                    return self.record_fail(
                        f"Battery is at {battery}% (limit: {self.minimum_battery}%)"
                    )
                return self.record_success(
                    f"Battery is at {battery}% (limit: {self.minimum_battery}%)"
                )
        return self.record_fail(
            f"Could not find {self.device_type} named {self.device_name}"
        )



"""
Gmirror array checks for simplemonitor.
"""

import subprocess

from .monitor import Monitor, register


@register
class MonitorGmirrorStatus(Monitor):
    """
    Check gmirror status for specified device.
    """

    monitor_type = "gmirror_status"

    def __init__(self, name: str, config_options: dict) -> None:
        super().__init__(name, config_options)
        self.array_device = self.get_config_option(
            "array_device", required_type="str", required=True
        )
        self.expected_disks = self.get_config_option(
            "expected_disks", required_type="int", required=True
        )

    def run_test(self) -> bool:
        """
        Run `gmirror status` for the specified device. Keep the logic simpler
        by requiring each device and the # of expected disks to be specified
        separately.
        """

        run_cmd = ["gmirror", "status", "-gs", self.array_device]
        try:
            result = subprocess.run(run_cmd, capture_output=True, check=True)
        except subprocess.CalledProcessError:
            return self.record_fail("gmirror command failed")

        status_lines = result.stdout.decode("utf-8").rstrip("\n").split("\n")

        # Status should be same for the array, so just grab first line
        status = status_lines[0].split()[1]

        # Infer # of disks based on the number of lines
        disk_count = len(status_lines)

        msg = f"Array {self.array_device} is in state {status} with {disk_count} disks"

        if status == "COMPLETE" and disk_count == self.expected_disks:
            return self.record_success(msg)

        return self.record_fail(msg)

    def describe(self) -> str:
        return f"Check RAID status of {self.array_device}."
